from cachingalgo.full_observation.single_cache import LU


class LU2(LU):
    def __init__(self, cache_sizes, **kwargs):
        """
        cache_sizes - array of cache size - length = 2
        """
        # Inherits the methods and attributes from LU Class
        super().__init__(**kwargs, cache_size=None)
        
        self.cache_size1 = cache_sizes[0]
        self.cache_size2 = cache_sizes[1]
        
        # self.fetch stores the time at which cache elements are fetched
        self.fetchtime1 = {}
        self.fetchtime2 = {}
    
    def cache1_update(self, req, time):
        """
        Updates the cache 1
        req: Request
        time: Request arrival time
        """
        vk = {}
        if len(self.fetchtime2) > 0:
            for i in self.fetchtime2:
                vk[i] = self.prob[i]*(self.fetchtime2[i]+self.F[i] - (time+1))

            maxindk = max(vk, key=vk.get) # Gives the key with maximum value

            if self.prob[req]*self.F[req] < vk[maxindk]:
                freshtime = self.fetchtime2.pop(maxindk)
                self.fetchtime1[maxindk] = freshtime
                self.fetchtime1.pop(req)
                self.fetchtime2[req] = time + 1
            else:
                self.fetchtime1[req] = time + 1
        else:
            self.fetchtime1[req] = time + 1
    
    def cache2_update(self, req, time):
        """
        Updates the cache 2
        req: Request
        time: Request arrival time
        """
        vj = {}
        for i in self.fetchtime1:
            vj[i] = self.prob[i]*(self.fetchtime1[i] + self.F[i] - (time+1))

        minindj = min(vj, key=vj.get) # Gives the key with minimum value
        if self.prob[req]*self.F[req] > vj[minindj]:
            freshtime = self.fetchtime1.pop(minindj)
            self.fetchtime2[minindj] = freshtime
            self.fetchtime2.pop(req)
            self.fetchtime1[req] = time + 1
        else:
            self.fetchtime2[req] = time + 1
    
    def cache12_update(self, req, time):
        """
        Updates the cache 1 and 2 simultaneously
        req: Request
        time: Request arrival time
        """
        vj = {} # Cache 1
        vk = {} # Cache 2

        for i in self.fetchtime1:
            vj[i] = self.prob[i]*(self.fetchtime1[i]+self.F[i] - (time+1))

        for i in self.fetchtime2:
            vk[i] = self.prob[i]*(self.fetchtime2[i]+self.F[i] - (time+1))

        minindj = min(vj, key=vj.get) # Gives the key with minimum value of cache 1
        minindk = min(vk, key=vk.get) # Gives the key with minimum value of cache 2

        if (self.prob[req]*self.F[req]) > vj[minindj]:
            if vj[minindj] >= vk[minindk]:
                freshtime = self.fetchtime1.pop(minindj)
                self.fetchtime2[minindj] = freshtime
                self.fetchtime2.pop(minindk)
                self.fetchtime1[req] = time + 1
            elif vk[minindk] > (self.prob[req]*self.F[req]):
                freshtime = self.fetchtime2.pop(minindk)
                self.fetchtime1[minindk] = freshtime
                self.fetchtime1.pop(minindj)
                self.fetchtime2[req] = time + 1
            else:
                self.fetchtime1.pop(minindj)
                self.fetchtime1[req] = time + 1
        elif (self.prob[req]*self.F[req]) >= vk[minindk]:
            self.fetchtime2.pop(minindk)
            self.fetchtime2[req] = time + 1
                    
    def currcache(self, req, time, ithreq = None):
        """
        Calculates the current cache
        req: Request
        time: Request arrival time
        ithreq: no. of request algorithm processed so far
        Returns: cache, cache_hit and miss_type
        """
        
        hit1 = 0 # hit '1' or miss '0' for cache 1
        hit2 = 0 # hit '1' or miss '0' for cache 2
        c2pass = 0 # Whether the request goes to cache 2
        
        if req >= self.L:
            raise Exception("The Requested item is not in the library")
        
        # Updates the popularity
        if self.calpop:    
            self.popularity_update(req, ithreq)

        # Check whether Cache1 is full or not
        if len(self.fetchtime1) < self.cache_size1:
            if self.method == 'lfulite' and req not in self.prob:
                self.arr[req] = [ithreq+1, 1] # storing the items that are in start of the cache
            self.fetchtime1[req] = time + 1
        
        # Serving request in Cache1
        elif req in self.fetchtime1:
            if self.fetchtime1[req] + self.F[req] >= (time + 1):
                hit1 = 1
            else:                
                self.cache1_update(req, time)
        
        # Check whether Cache2 is full or not
        elif len(self.fetchtime2) < self.cache_size2:
            c2pass = 1
            if self.method == 'lfulite' and req not in self.prob:
                self.arr[req] = [ithreq+1, 1] # storing the items that are in start of the cache
            self.fetchtime2[req] = time + 1
        
        # Serving the request from the Cache2
        elif req in self.fetchtime2:
            c2pass = 1
            if self.fetchtime2[req] + self.F[req] >= (time+1):
                hit2 = 1
            else:
                self.cache2_update(req, time)
        
        # Serving the request from the library
        else:
            c2pass = 1
            if self.method == 'lfu' or ((self.method == 'lfulite')^(req not in self.prob)):
                self.cache12_update(req, time)
                    
        return {'cache':[[*self.fetchtime1.keys()],[*self.fetchtime2.keys()]], 'cache_hit':[hit1,hit2], 'c2pass':c2pass}
    





from collections import deque
import numpy as np
import time
import pickle
from pathlib import Path

class StaticZipf:
    def __init__(self, L, a):
        """
        L is the Library Size
        a is the shape parameter of the Zipf distribution
        """
        self.L = L
        self.a = a
        self.prob = self.zipf(L, a)

    def req_gen(self, count):
        """
        Generates the 'count' no. of requests according to self.prob
        """
        items = list(range(self.L))
        req = np.random.choice(items, p = self.prob, size = count)
        return req

    def zipf(self, n, a):
        """
        n is the size
        a is the shape factor
        pdf of X, fx(X) = c/(x^a) where c = summation of (1/i)^a where i runs from 1 to n
        To get the top one element one has to substitute x = 1 and vice versa
        """
        x = sum([(1/i)**a for i in range(1,n+1)])
        y = [None]*n
        for i in range(n):
            y[i] = (1/(i+1)**a)*(1/x)
        return np.array(y)

    def rotatetop(self, y, top, step, direction):
        """
        y is the array to cyclically shift the first top elements by step.
        step: amount which the we have to shift
        direction: +1/-1 right or left
        """
        s = deque(y[:top])
        s.rotate(direction*step)
        return np.array(list(s)+list(y[top:]))



def szipf(L=100, count=10_000, a=1, silent=True):
    """
    Generates the static zipf requests

    Refer to the Static Zipf Class to understand variables
    Returns the prob distribution and generated requests dictionary
    """

    start = time.time()

    szipf = StaticZipf(L=L, a=a)
    prob = szipf.prob
    req = szipf.req_gen(count)

    if not silent:
        print(f'TimeTaken:{time.time() - start} sec')

    return {'prob':prob, 'req':req}


def dzipf(L=100, count=10_000, req_step=1000, window=50, top=10, cache_size=10, a=1, silent=True):
    """
    Generates the dynamic zipf requests

    Refer to the Static Zipf Class to understand variables
    req_step: no. of requests after which popularity changes
    window: no. of most popular items for which the popularities are changed
    top: the no. of item's popularities are moved to left or right
    Returns the generated requests and optimal cache
    """

    # Input checking
    if not np.all([window < L, top < window, count % req_step == 0, top < L]):
        raise ValueError("Input arguments are not valid")

    start = time.time()

    count_step = int(count/req_step)

    szipf = StaticZipf(L=L, a=a)
    optimcache = np.zeros((count_step*cache_size)).astype(int)
    req = np.zeros((count,)).astype(int)
    for i in range(count_step):
        optimcache[cache_size*i:cache_size*(i+1)] = np.argsort(szipf.prob)[::-1][:cache_size]
        req[req_step*i:req_step*(i+1)] = szipf.req_gen(req_step)
        szipf.prob = szipf.rotatetop(szipf.prob, window, top, 1)

    if not silent:
        print(f'TimeTaken:{time.time() - start} sec')

    return {'optim_cache':optimcache, 'req':req}

# Netflix Data is collected between 1998 and 2005
# Library Size - 17, 770
# Requests logs processed - 10_04_80_507
def netflix(count=10_000, silent=True):
    """
    Loads the popularity profile from netflixdata file and generates the requests

    count: No. of requests to be generated
    Returns the prob distribution and generated requests dictionary
    """
    curr_path = Path(__file__).resolve().parents[1]
    data_path = Path("./data/netflixdata.pkl")

    with open(curr_path / data_path,'rb') as f:
        popularity_dic = pickle.load(f)

    items = list(popularity_dic.keys())
    prob = list(popularity_dic.values())

    start = time.time()
    # Random sampling
    req = np.random.choice(items, p = prob, size = count)
    if not silent:
        print(f'TimeTaken:{time.time()-start}sec')
    return {'prob':prob, 'req':req}


# Library Size - 1,61,085
def youtube(count=10_000, silent=True):
    """
    Loads the popularity profile from youtube data file and generates the requests

    count: No. of requests to be generated
    Returns the prob distribution and generated requests dictionary
    """
    curr_path = Path(__file__).resolve().parents[1]
    data_path = Path("./data/youtubedata.pkl")

    start = time.time()
    with open(curr_path / data_path,'rb') as f:
        prob = pickle.load(f)

    items = list(range(prob.shape[0]))
    # Random Sampling
    req = np.random.choice(items, p = prob, size = count)
    if not silent:
        print(f'TimeTaken:{time.time()-start}sec')
    return {'prob':prob, 'req':req}



import numpy as np
import random
from collections import deque
import math

# Caching Bandit Marginal Posterior Sampling
class CBMPS:
    def __init__(self, L, cache_size):
        """
        L: Library Size
        cache_size: Cache Size
        """
        # Index 0 is alpha, Index 1 is beta of the beta distribution and Index 2 is Popularity.
        self.param = {i:[1,1,0] for i in range(L)}
        self.L = L
        self.cache_size = cache_size
        self.cache = random.sample(range(0,self.L),self.cache_size)

    def currcache(self, Return = False):
        """
        Calculates the current cache
        Return: True - returns the currcache or False - void return
        """
        # Generating the samples from the beta distribution for the items in the library.
        for i in self.param:
            #.item() returns the element in the single element numpy array
            self.param[i][2] = np.random.beta(self.param[i][0], self.param[i][1],1).item()

        #Sorting in decreasing order according to popularity distribution.
        sort = [k for k,_ in sorted(self.param.items(), key = lambda item: item[1][2], reverse = True)]

        #Updating the Current Cache
        self.cache = sort[:self.cache_size]

        if Return:
            return self.cache


    def update(self, req):
        """
        Updates the paramters of the beta distribution
        req: Request
        """

        #Updates only if the element is in the current cache
        if req in self.cache:
            #incrementing the alpha of ele
            self.param[req][0] += 1

            #incrementing the beta of all items in the currcache except for ele
            for i in self.cache:
                if i != req:
                    self.param[i][1] += 1

# Caching Bandit Structural Information
class CBSI:
    def __init__(self, L, cache_size, mu_c, delta):
        """
        L: Library Size
        cache_size: Cache Size
        mu_c: Probability of most popular Cth library item
        delta: Difference b/n probabilities of Cth and C+1 th most popular library items
        """
        self.L = L
        self.cache_size = cache_size
        # Index 0 is alpha, Index 1 is beta
        self.param = {i:[0, 0] for i in range(self.L)}
        # Probability of item
        self.prob = {i: 1/self.L for i in range(self.L)}
        # Cache at time t
        self.cache = random.sample(range(self.L), self.cache_size)
        self.mu_c = mu_c
        self.delta = delta
        # Library used for sampling
        self.Lib = self.prob

    def compute_A(self):
        """
        Calculates the dictionary A along with the probability according to mu_c and delta
        Returns A
        """
        A = {}
        for i in self.Lib:
            if self.Lib[i] >= (self.mu_c - (self.delta / 2)):
                A[i] = self.Lib[i]
        return A

    def sampling(self, pr, number):
        """
        Calculates the sampled set of items from the probability dictionary
        Return Set B
        """
        indices = list(pr.keys())
        probabilities = np.hstack(list(pr.values()))
        # B contains unique items
        B = np.random.choice(indices,p = probabilities, size = number, replace = False)
        return set(B)

    def currcache(self, Return = False):
        """
        Calculates the current cache
        Return: True - returns the currcache or False - void return
        """
        # Computing A with along with probabilities
        A = self.compute_A()

        if len(A) >= self.cache_size:
            # Sorting in decreasing order according to popularity distribution.
            sort = [k for k,_ in sorted(A.items(), key = lambda item: item[1], reverse = True)]

            # Updating current cache
            self.cache = sort[:self.cache_size]

        else:
            # Calculating the prob distribution (potential function) for the elements in the Lib other than in set A
            setA = set(A.keys())
            p = {}

            for i in self.Lib:
                if i not in A:
                    p[i] = 1/((self.mu_c - self.Lib[i])**2)

            c = sum(p.values())

            pdf = {k: v/c for k,v in p.items()}

            # Returns the setB which are sampled from the items that are not in setA according to prob
            setB = self.sampling(pdf, self.cache_size - len(A))

            self.cache = list(setA.union(setB))

        if Return:
            return self.cache

    def update(self, req):
        """
        Updates the parameters of the algorithm
        """

        #Checks whether request present in cache
        if req in self.cache:

            #Increase the alpha for the req
            self.param[req][0] += 1

            #Increase the beta for the items other than req
            for ele in self.cache:
                if ele != req:
                    self.param[ele][1] += 1

                #Updates the popularity distribution
                self.Lib[ele] = self.param[ele][0] / (self.param[ele][0] + self.param[ele][1])


def calculate_delta(prob, cache_size):
    """
    Calculates the delta which mu_c - mu_c+1, and mu_c
    Returns delta and mu_c
    """
    #calculates the delta = mu_c - mu_c+1
    prob_c_plus1 = np.sort(prob)[-(cache_size+1):][::-1]
    mu_c = prob_c_plus1[cache_size-1]
    mu_c_plus1 = prob_c_plus1[cache_size]
    delta = mu_c - mu_c_plus1
    return delta, mu_c


class CBSILite(CBSI):
    def __init__(self, win=None, **kwargs):
        """
        win: Window Size
        """
        # Inherits the methods and attributes from CBSI Class
        super().__init__(**kwargs)

        if win == None:
            win = int(self.cache_size*self.cache_size*math.log(self.L))

        # deque internally uses double linked list
        self.q = deque(i for i in np.random.randint(self.L, size = win)) # Randomly intialised deque of size window upto L integers

        # Ocuurences of elements in the deque are intialised
        self.dic = {}
        for i in self.q:
            if i in self.dic:
                self.dic[i] +=1
            else:
                self.dic[i] = 1

        #Counter_bank
        self.Lib = {}

    def deq_update(self, req):
        """
        Updates the deque q
        req: Request
        """
        # appends the new request to the deque
        self.q.appendleft(req)
        # removes the last request from the deque
        rem = self.q.pop()
        # Decreasing the count or removing the request
        if self.dic[rem] == 1:
            del self.dic[rem]
        else:
            self.dic[rem] -= 1
        # Increasing the count or appending the request
        if req in self.dic:
            self.dic[req] += 1
        else:
            self.dic[req] = 1

    def freq_top(self):
        """
        Returns the most frequently occuring items in the window
        """
        return [k for k, _ in sorted(self.dic.items(), key = lambda x: x[1], reverse = True)[:self.cache_size]]

    def counterbank_update(self, req):
        """
        Updates the Counter Bank
        req: Request
        """
        self.deq_update(req)
        topfreq = self.freq_top()

        for i in topfreq:
            if i not in self.Lib:
                self.Lib[i] = self.prob[i]



from scipy.optimize import minimize_scalar, minimize
import numpy as np
import random

# Cost Minimisation with Same Refresh Rate
class CMSR:
    def __init__(self, cache_size, beta, z, lambda_param, L, Cost, prob=[], mu_hat=[], C_hat=[]):
        """
        Cost : [C_f, C_ca, C_ch, C_o] = [fetching, additional caching, checking, constant]
        beta : Arrival Rate
        cache_size : Size of the Cache
        lambda_param : lambda parameter
        z: order of the lambda_ (Refresh rates of the library gets decreased)
        prob: Popularity profile of the contents in the library
        mu_hat: optimal refresh rate
        C_hat: optimal cache elements
        """
        self.prob = prob
        self.L = L
        self.cost = Cost
        self.cache_size = cache_size
        self.beta = beta
        self.lambda_param = lambda_param
        self.lambda_ = np.array([self.lambda_param/(i)**z for i in range(1,self.L+1)],dtype=np.float64) # Refresh rates of the contents
        self.C_hat = np.array(C_hat)                              # Current Cache
        self.C_old =  np.argsort(self.prob)[::-1][:self.cache_size]# Old cache
        self.firsttime = True
        self.mu_hat = self.minimize_cost()                       # Checking Rate of the cache
        self.firsttime = False
        self.cache_update()                                    # Calculated cache

    # Average System Cost
    def avg_sys_cost(self, rr):
        """
        Calculates the average system cost
        rr: Refresh rate of the cache
        Returns the average system cost
        """
        eps = 1e-15
        # if we are
        if not self.firsttime:
            d = self.C_hat
        else:
            d = self.C_old

        try:
            return self.beta*self.cost[0] + \
                        len(self.C_hat)*rr*self.cost[2] + \
                        rr*self.cost[1]*np.sum((self.lambda_[d])/(self.lambda_[d] + rr)) + \
                        self.beta*np.sum(self.prob[d]*((self.lambda_[d]*self.cost[-1]/(rr+eps))-self.cost[0]))
        # if cache contains no elements
        except:
            return self.beta*self.cost[0]

    def minimize_cost(self):
        """
        Convex Constrained Minimisation of the average system Cost
        Returns the optimal refresh rate of the cache
        """
        return minimize_scalar(self.avg_sys_cost, bounds = (0, None)).x

    def marg_cost(self):
        """
        Calculates the marginal cost vector of the items in the library
        Returns the marginal cost vector
        """
        eps = 1e-15
        return self.mu_hat*self.cost[2] +\
            ((self.mu_hat*self.lambda_)/(self.lambda_ + self.mu_hat))*self.cost[1] + \
            self.beta*self.prob*(((self.cost[3]*self.lambda_)/(self.mu_hat+eps))-self.cost[0])

    def cache_update(self):
        """
        Updates the Cache according to the marginal cost
        """
        mc = self.marg_cost()
        sorted_mc = np.argsort(mc[mc<0])
        if len(sorted_mc) > self.cache_size:
            self.C_hat = sorted_mc[:self.cache_size]
        else:
            self.C_hat = sorted_mc

    def refreshrate_cache(self):
        """
        Calculates the optimal refresh rate of the cache and optimal cache after their convergence
        """
        while(not np.array_equal(self.C_hat, self.C_old)):
            self.C_old = self.C_hat
            self.mu_hat = self.minimize_cost()
            self.cache_update()

    def currcache(self):
        """
        Calculates and returns the current cache
        """
        self.refreshrate_cache()
        return self.C_hat

# Cost Minimization with Different Refresh Rates
# Variants of CMDR- CMDR: Popularity is given
#                   CMDRP: Popularity is not given and cache update according to popularity
#                   CMDRM: Popularity is not given and cache update according to marginal cost 
class CMDR:
    def __init__(self, cache_size, beta, z, lambda_param, L, Cost, cache_update = 'marg', seed = 7, prob = [], C_hat=[]):
        """
        Cost : [C_f, C_ca, C_ch, C_o] = [fetching, additional caching, checking, constant]
        beta : Arrival Rate
        cache_size : Size of the Cache
        lambda_param : lambda parameter
        prob: probability profile of the items in the library
        z: order of the lambda_ (Refresh rates of the library gets decreased)
        cache_update: 'marg' - updates the cache according to the which has the more negative marginal cost
                    'pop'- updates the cache according to the most popular items in the library.
        C_hat: optimal cache of length cache_size
        seed: seed for numpy and random
        """
        np.random.seed(seed)
        random.seed(seed)
        self.upd = cache_update
        self.L = L
        self.cost = Cost
        self.cache_size = cache_size
        self.beta = beta
        self.lambda_param = lambda_param
        self.lambda_ = np.array([self.lambda_param/(i)**z for i in range(1,self.L+1)],dtype=np.float64) # Refresh rates of the contents
        self.C_hat = np.array([], dtype='int') if len(C_hat) == 0 else C_hat
        self.mu_hat = np.array([random.uniform(0,i) for i in self.lambda_])
        # Setting Bounds for the cost minimisation
        self.bounds = [(0, i) for i in self.lambda_]
        if len(prob) == 0:
            self.prob = np.zeros((L,))
            self.arr = np.zeros((L,)) # arary Counter
            self.calpop = True
        else:
            self.prob = prob
            self.calpop = False
            self.minimize_cost()


    # Average System Cost
    def avg_sys_cost(self, rr):
        """
        Calculates the average system cost
        rr: Refresh rates of the cache elements
        Returns the average system cost
        """
        eps = 1e-15

        # For the empty cache it returns self.beta*self.cost[0]
        try:
            return self.beta*self.cost[0] + \
                    np.sum(rr[self.C_hat])*self.cost[2] + \
                    self.cost[1]*np.sum((rr[self.C_hat]*self.lambda_[self.C_hat])/(self.lambda_[self.C_hat] + rr[self.C_hat])) + \
                    self.beta*np.sum(self.prob[self.C_hat]*((self.lambda_[self.C_hat]*self.cost[-1]/(rr[self.C_hat]+eps))-self.cost[0]))
        except:
            return self.beta*self.cost[0]

    # Minimising the System Cost for the Cache size equal to the library size i.e., infinite Cache
    def avg_sys_cost_inf(self, rr):
        """
        Calculates the average system cost if cache consists of entire library
        rr: Refresh rates of the entire library
        Returns the average system cost
        """
        eps = 1e-15
        return self.beta*self.cost[0] + \
                        np.sum(rr)*self.cost[2] + \
                        self.cost[1]*np.sum((rr*self.lambda_)/(self.lambda_ + rr)) + \
                        self.beta*np.sum(self.prob*((self.lambda_*self.cost[-1]/(rr+eps))-self.cost[0]))

    # Convex Constrained Minimisation

    def minimize_cost(self):
        """
        Convex Constrained Minimisation of the average system Cost
        Returns the optimal refresh rates of the cache
        """
        # x0 - initial values for the cost minimization
        return minimize(self.avg_sys_cost, x0=self.mu_hat[self.C_hat], bounds=[self.bounds[i] for i in self.C_hat]).x


    # Caluclates the marginal cost vector and returns it
    def marg_cost(self):
        """
        Calculates the marginal cost vector of the items in the library
        Returns the marginal cost vector
        """
        eps = 1e-15
        return self.mu_hat*self.cost[2] +\
            ((self.mu_hat*self.lambda_)/(self.lambda_ + self.mu_hat))*self.cost[1] + \
            self.beta*self.prob*(((self.cost[3]*self.lambda_)/(self.mu_hat+eps))-self.cost[0])

    def cache_update(self, req):
        """
        Updates the cache
        req: Request
        """
        if len(self.C_hat) < self.cache_size:
            if req not in self.C_hat:
                self.C_hat = np.append(self.C_hat, req)
        else:
            self.mu_hat[self.C_hat] = self.minimize_cost()
            if self.upd == 'marg':
                if req not in self.C_hat:
                    mc = self.marg_cost()
                    item_rem = self.C_hat[np.argmax(mc[self.C_hat])]
                    if mc[item_rem] > mc[req]:
                        self.C_hat = np.delete(self.C_hat, np.where(self.C_hat == item_rem))
                        self.C_hat = np.append(self.C_hat, req)
            elif self.upd == 'pop':
                self.C_hat = np.argsort(self.prob)[::-1][:self.cache_size]


    def counter_update(self, req):
        """
        Increases the counter of the library items
        req: Request
        """
        self.arr[req] += 1

    # Calculates the popularity distribution
    def currcache(self, req):
        """
        Calculates and Returns the current cache
        """
        if self.calpop:
            self.counter_update(req)
            self.prob = self.arr/np.sum(self.arr)
            self.cache_update(req)
        return self.C_hat



import random
from cachingalgo.full_observation.single_cache import LRU


# LRUm algorithm consists of v virtual caches, f-v caches of total size m.
class LRUm(LRU):
    def __init__(self, size, f, v, L):
        """
        size: vector of length f which specifies the length of each list
               If length = 1 and f > 1, all the lists have same length
        f: Total no. of lists used in the algorithm
        v: No. of lists used as virtual Caches.
        L: Library size
        """

        self.f = f
        # No. of lists used as Cache
        self.cnum = f-v
        self.L = L

        # Uses LRU if LRUm uses one list
        if f == 1:
            super().__init__(cache_size=size, L=L)
        else:
            if isinstance(size, list) and len(size) == f:
                self.size = size
            else:
                self.size = [size]*f

            # self.part is to make sure that at the start, each of lists are unique
            self.part = self.L//self.f
            self.collection = []

            #Randomly initializes the f lists
            for itr, length in enumerate(self.size):
                start = int(itr*self.part)
                end = int((itr+1)*self.part)
                self.collection.append(random.sample(range(start, end), int(length)))

    def index_finder(self, req, full_search = False):
        """
        Finds the index of request in the f lists
        req: Request
        full_search: True - searches all list or False - searches only cache excluding virtual caches
        Returns request's position in list and it's corresponding list index. returns -1 if not found
        """
        if full_search:
            part = self.cnum
        else:
            part = self.f
        index = -1
        listnum = -1

        # Searches from the last list
        for i, list_ in enumerate(self.collection[-part:]):
            if req in list_:
                index = list_.index(req)
                listnum = i
        return index, listnum

    def update(self, req):
        """
        Updates the f lists
        req: Request
        """

        # If the algorithm has more lists
        if self.f != 1:
            # Gives the index, list number of the req
            index, listnum = self.index_finder(req)

            if index != -1:
                if listnum < self.f-1:
                    # moves the req to the index 0 of listnum + 1 and listnum+1 last element
                    # is moved to the index 0 of listnum
                    self.collection[listnum].remove(req)
                    temp = self.collection[listnum+1].pop()
                    self.collection[listnum] = [temp] + self.collection[listnum]
                    self.collection[listnum+1] = [req] + self.collection[listnum+1]

                else:
                    # moves the req to the index 0
                    temp = self.collection[listnum].pop(index)
                    self.collection[listnum] = [temp] + self.collection[listnum]

            # adds the ele at index 0 of list 1 and discards the last element of that list.
            else:
                self.collection[0].pop()
                self.collection[0] = [req] + self.collection[0]

        else:
            super().update(req)

    def __contains__(self, req):
        """
        Magic method to use "in" keyword
        req: request
        """
        if self.f != 1:
            index, _ = self.index_finder(req, True)
            if index != -1:
                return True
            else:
                return False

        else:
            if req in self.cache:
                return True
            else:
                return False

    def currcache(self):
        """
        Returns the current cache
        """

        if self.f != 1:
            return self.collection[-self.cnum:]
        else:
            return self.cache

# fLRU algorithm consists of f-1 virtual caches and 1 cache
class fLRU:
    def __init__(self, f, size, L):
        """
        f: total no. of caches
        size: vector of length f which specifies the length of each list
               If length = 1 and f > 1, all the lists have same length
        L: Library Size
        """
        self.f = f
        self.L = L
        if isinstance(size, list) and len(size) == f:
            self.size = size
        else:
            self.size = [size]*f

        self.collection = []

        #Randomly intialiases the f lists
        for length in self.size:
            self.collection.append(random.sample(range(1, L), length))

    def update(self, req):
        """
        Updates the f lists
        req: Request
        """
        found = False
        for i in range(self.f-1, -1, -1):
            currlist = self.collection[i]
            prevlist = self.collection[i-1] if i > 0 else None

            # if the req is present in the ith list, move it to the first position
            if req in currlist:
                currlist.remove(req)
                self.collection[i] = [req] + currlist
                found = True

            # if the req is present in the i-1th list but not in ith list, move it
            # to the first position of the ith list and discard the last item of the ith list.
            elif prevlist is not None and req in prevlist:
                _ = currlist.pop()
                self.collection[i] = [req] + currlist
                found = True

        # if the req is not in the collection of lists then insert it in the first list.
        if not found:
            _ = self.collection[0].pop()
            self.collection[0] = [req] + self.collection[0]

    def __contains__(self, req):
        """
        Magic method to use "in" keyword
        req: request
        """
        if req in self.collection[-1]:
            return True
        else:
            return False

    def currcache(self):
        """
        Returns the current cache
        """
        return self.collection[-1]



import numpy as np
from collections import deque
import math
import random

# Least Frequently Used
class LFU:
    def __init__(self, L, cache_size):
        """
        L : Library size
        cache_size: Size of the Cache
        """
        self.arr = np.zeros((L,)) #intialises the counters array of size = library size
        self.cache_size = cache_size
        self.prob = np.zeros((L,)) # Stores the probability distribution of the library items

    def update(self, req):
        """
        Updates the counter of library items
        req: request
        """
        self.arr[req] += 1

    def currcache(self, Return = True, exclude = []):
        """
        To find the cache using LFU
        exclude: items that are to be excluded from the library while finding cache
        Return: True - returns the current cache or False - doesn't return
        Returns the current cache according to the popularity
        """
        distrib = self.arr/np.sum(self.arr)
        self.prob = distrib
        if len(exclude) != 0 :
            distrib[exclude] = 0
        if Return:
            return np.argsort(distrib)[::-1][:self.cache_size]

    def popularity(self):
        """
        Returns the popularity of the items in the Library
        """
        return self.prob

    def counters_used(self):
        """
        Returns no. of counters used till now
        """
        return np.count_nonzero(self.arr)

# Window LFU
class WLFU:
    def __init__(self, L, cache_size, window=None, F=[]):
        """
        window: Size of the window
        L: Library size
        F: Freshness constraints of the library items and size = L
        cache_size: Cache Size
        """
        if window == None:
            self.window = int(cache_size*cache_size*math.log(L))
        else:
            self.window = window
        self.L = L
        self.cache_size = cache_size

        # Checking whether we need to take F into consideration
        if len(F) == 0:
            self.state = False
        else:
            self.state = True
            # divided by window to normalize the Freshness constriants
            self.F = np.array(F)/self.window


        #deque internally uses double linked list
        self.q = deque(i for i in np.random.randint(L, size = self.window)) # Randomly intialised deque of size window upto L integers
        self.dic = {}

        # Random initialization of the deque
        if not self.state:
            for i in self.q:
                if i in self.dic:
                    self.dic[i] += 1
                else:
                    self.dic[i] = 1
        else:
            for i in self.q:
                if i in self.dic:
                    self.dic[i] += self.F[i]
                else:
                    self.dic[i] = self.F[i]

    def update(self, req):
        """
        Updates the counter deque used in WLFU
        req: request
        """

        #appends the new request to the deque
        self.q.appendleft(req)

        #removes the last request from the deque
        rem = self.q.pop()

        #Decreasing the count or removing the request
        if not self.state:
            if self.dic[rem] == 1:
                del self.dic[rem]
            else:
                self.dic[rem] -= 1
        else:
            if self.dic[rem] == self.F[rem]:
                del self.dic[rem]
            else:
                self.dic[rem] -= self.F[rem]

        #Increasing the count or appending the request
        if not self.state:
            if req in self.dic:
                self.dic[req] += 1
            else:
                self.dic[req] = 1
        else:
            if req in self.dic:
                self.dic[req] += self.F[req]
            else:
                self.dic[req] = self.F[req]

    def currcache(self):
        """
        to find the cache according frequency of items in window
        Returns the current cache
        """

        sort_arr = dict(sorted(self.dic.items(), key= lambda x:x[1], reverse = True)[:self.cache_size]).keys()
        return np.array(list(sort_arr))


class LFULite:
    def __init__(self, L, cache_size, window=None, F=[]):
        """
        window: window size of WLFU
        L: Library Size
        cache_size: Cache Size
        F: Freshness Constraints of the library items
        """
        self.L = L
        self.cache_size = cache_size
        self.counterbank = {} # Consists of key as video ID and value as list of [time, number of occurences]
        self.wlfu = WLFU(L=L, cache_size=cache_size, F=F, window=window)
        self.prob = {} # Item id and it's probability

    def update(self, req, ithreq, wlfu = True):
        """
        Update the counterbank and counter in WLFU
        req: request
        wlfu: whether to update the wlfu counter bank or not
        ithreq: no. of request algorithm processed so far
        """
        #Update the counter of the WLFU
        if wlfu:
            self.wlfu.update(req)

        #most frequent items in the window
        currtop = self.wlfu.currcache()

        #If present in the counterbank it increases the count.
        if req in self.counterbank:
            self.counterbank[req][1] += 1

        #Appending the elements that are not in counterbank
        for j in list(currtop):
            if j not in self.counterbank:
                self.counterbank[j] = [ithreq+1,1]

    def currcache(self, ithreq, Return = True, exclude = []):
        """
        to find the cache according to LFULite
        ithreq: how many request that are given to the algorithm till now
        exclude: items to be excluded while calculating the cache
        Return: True - returns the current cache or False - doesn't return
        Returns the current cache
        """
        if ithreq > 0:
            arr = {i:0 for i in self.counterbank}
            for k in arr:
                # Caluclating the popularity distribution of the elements that we added before this time.
                if self.counterbank[k][0] != (ithreq+1):
                    # Counting the appearnces after it is been added to the counterbank
                    arr[k] = (self.counterbank[k][1]-1)/((ithreq+1)-self.counterbank[k][0])
                self.prob = arr

            # For excluding some library item's in the cache
            if len(exclude) != 0:
                # it will assign 0
                arr.update(dict.fromkeys(exclude, 0))

            sort_arr = dict(sorted(arr.items(), key= lambda x:x[1], reverse = True)[:self.cache_size]).keys()
            if Return:
                return np.array(list(sort_arr))
        else:
            if Return:
                return np.array(list(self.counterbank.keys()))

    def popularity(self):
        "Return the popularity of the items in the counterbank"
        return self.prob

    def counters_used(self):
        """
        Returns no. of counters used till now
        """
        return len(self.counterbank)

class CountSketch:
    def __init__(self, l, b, L):
        """
        t: No. of random functions for each s and h
        b: No. of objects that h has to map from L
        L: Library Size
        """
        self.l = l
        self.b = b
        self.L = L
        self.h, self.s = self.createmap()
        self.cs = np.zeros((l,b))
        # Cache is of size C but as we are using zipf with parameter 1. C = b
        self.cache = []

    def randomassign(self, num, bi = False):
        """
        Generates random mapping value for mapping function
        num: no. of different mapping values
        bi: True - +1 or -1 and False - 1 to num (both inclusive)
        Returns generated random number in the given range
        """
        if bi:
            n = random.random()
            if(n > 0.5):
                return 1
            else:
                return -1
        else:
            return int(random.randint(1, num))

    def createmap(self):
        """
        Creates l number of h and s mapping functions for each library item.
        Return h and s mapping functions
        """
        h = {i:[] for i in range(self.L)}
        s = {i:[] for i in range(self.L)}

        # Creating l number of h functions
        for _ in range(self.l):
            for i in range(self.L):
                n = self.randomassign(self.b)
                h[i].append(n)

        # Creating l number of s functions
        for _ in range(self.l):
            for i in range(self.L):
                n = self.randomassign(2, True)
                s[i].append(n)

        return h,s

    def update(self, req):
        """
        Updates the cs counters which is a t*b array
        req: Request
        """
        for i in range(self.l):
            self.cs[i,self.h[req][i]-1] += self.s[req][i]

    def estimate(self, req):
        """
        Calculates the estimate of the requested item by finding median hi[r(t)]*si[r(t)]
        req: Request
        """
        med = np.zeros((self.l,))
        for i in range(self.l):
            med[i] = self.cs[i,self.h[req][i]-1]*self.s[req][i]
        return np.median(med)

    def min_est(self):
        """
        Calculates the minimum estimate of the cache and it's position
        Returns position and minimum value
        """
        arr = np.zeros((len(self.cache),))
        for i,j in enumerate(self.cache):
            arr[i] = self.estimate(j)
        index = np.argmin(arr)

        return index, arr[index]

    def currcache(self, req):
        """
        Updates the cs counter and inserts the request into the cache
        req: Request
        Returns the current cache
        """
        self.update(req)
        # If length of cache is less than b and not in cache.
        if(len(self.cache)<self.b and req not in self.cache):
            # Adding the request to the cache
            self.cache.append(req)
        elif(req not in self.cache):
            # Finding the estimate of the request
            req_est = self.estimate(req)
            # Adding the new request and removing the min_est item from the cache
            ind, est = self.min_est()
            if req_est > est:
                self.cache[ind] = req

        return np.array(self.cache)

# Least Recently Used
class LRU:
    def __init__(self, cache_size, L):
        """
        cache_size: Cache Size
        L: Library Size
        """
        self.cache_size = cache_size
        self.L = L
        self.cache = random.sample(range(1,L+1), cache_size)

    def update(self, req):
        """
        Updates the elements in the cache
        req: request
        """
        currcache = self.cache

        # If present in the cache, bring the item to the begining of the list
        if req in self.cache:
            currcache.remove(req)
            self.cache = [req] + currcache
        # else remove the last element of the list and insert in the first position.
        else:
            _ = currcache.pop()
            currcache.insert(0, req)
            self.cache = currcache

    def __contains__(self, req):
        """
        Magic method to use "in" keyword
        req: request
        """
        if req in self.cache:
            return True
        else:
            return False

    def currcache(self):
        "Returns the current cache"
        return self.cache

# Least Useful: Refer LFU, LFULite and WLFU algorithms to understand.
class LU:
    def __init__(self, L, F, cache_size, arr = [], method = '', useF = False, freqtop=None, window=None):
        """
        L: Library Size
        F: Freshness constant array of size L
        cache_size: Size of the cache for the algorithm
        freqtop: No. of the most frequent elements wlfu needs to consider
        useF: use of F for WLFU
        window: window size of the WLFU
        arr: probability of library ites in case of LU algorithm
        method: 'lfu' or 'lfulite': Default is 'lfu' if arr is not given
        """
        self.F = {i:j for i, j in enumerate(F)}
        self.L = L
        self.cache_size = cache_size
        self.method = method

        # self.calpop determines whether we have to calculate popularity or not
        self.calpop = True if len(arr) == 0 else False

        # self.arr counter for the library items
        # self.prob contains the probabilty of items used by the algorithm
        if not self.calpop:
            self.prob = np.array(arr)
        elif self.method == 'lfu':
            self.arr = np.zeros((L,))
            self.prob = np.zeros((L,))
        elif self.method == 'lfulite': # LFU-Lite maintains the popularity of the items only in the counter bank
            self.arr = {} # Consists of key as video ID and value as list of [time, number of occurences]
            if useF:
                self.wlfu = WLFU(L=L, cache_size=freqtop, F=F, window=window)
            else:
                self.wlfu = WLFU(L=L, cache_size=freqtop, window=window)
            self.prob = {}
        # self.fetchtime stores the time at which cache elements are fetched
        self.fetchtime = {}

    def update(self, req, ithreq = None):
        """
        Updates the counters of the items used in the algorithm
        req: Request
        ithreq: no. of request algorithm processed so far
        """

        if self.method == 'lfu':
            self.arr[req] += 1
        elif self.method == 'lfulite':
            # finds the most frequent items according to WLFU
            self.wlfu.update(req)
            currtop = self.wlfu.currcache()
            if req in self.arr:
                self.arr[req][1] += 1

            for j in currtop:
                if j not in self.arr:
                    self.arr[j] = [ithreq+1, 1]

    def cache_update(self, req, time):
        """
        Updates the cache based on the request and it's arrival time
        req: Request
        time: arrival time of the request
        """
        v = {}

        for i in self.fetchtime:
            v[i] = self.prob[i]*(self.fetchtime[i]+self.F[i] - (time+1))

        minind = min(v, key=v.get) # Gives the key with minimum value

        if (self.prob[req]*self.F[req]) > v[minind]:
            self.fetchtime.pop(minind)
            self.fetchtime[req] = (time+1)

    def popularity_update(self, req, ithreq):
        """
        Calculates and updates the popularity of the items used in the algorithm
        req: Request
        ithreq: no. of request algorithm processed so far
        """
        self.update(req, ithreq = ithreq)
        if self.method == 'lfu':
            distrib = self.arr/np.sum(self.arr)
        elif self.method == 'lfulite':
            distrib = {i:0 for i in self.arr}
            for k in distrib:
                if self.arr[k][0] != (ithreq+1):
                    distrib[k] = (self.arr[k][1]-1)/((ithreq+1)-self.arr[k][0])
        self.prob = distrib

    def currcache(self, req, time, ithreq = None):
        """
        Calculates the current cache
        req: Request
        time: Request arrival time
        ithreq: no. of request algorithm processed so far
        Returns: cache, cache_hit and miss_type
        """
        hit = 0
        miss_type = -1 # Hit

        if req >= self.L:
            raise Exception("The request is not in the library")

        # Updates the popularity
        if self.calpop:
            self.popularity_update(req, ithreq)

        # Check whether cache is full or not
        if len(self.fetchtime) < self.cache_size:
                if self.method == 'lfulite' and req not in self.prob:
                    self.arr[req] = [ithreq+1, 1] # storing the items that are in start of the cache
                self.fetchtime[req] = time + 1
                miss_type = 1 # miss due to freshness constraint
        else:
            if req in self.fetchtime:
                if self.fetchtime[req] + self.F[req] >= (time + 1):
                    hit = 1
                else:
                    self.fetchtime[req] = (time+1)
                    miss_type = 1
            else:
                miss_type = 2 # miss due to not present in the cache

                if self.method == 'lfu' or ((self.method == 'lfulite')^(req not in self.prob)):
                    self.cache_update(req, time)

        return {'cache':list((self.fetchtime.keys())), 'cache_hit':hit, 'miss_type':miss_type}

    def popularity(self):
        "Return the popularity of the items used in the counterbank"
        return self.prob



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
      name='cachingalgo',
      packages = find_packages("cachingalgo"),
      version='0.0.2',
      description='Package consists of implementation of caching algorithms and request generation modules',
      author="Surya Krishna Guthikonda",
      author_email="suryakguthikonda@gmail.com",
      url="https://github.com/SuryaKrishna02/caching-algorithms",
      keywords = ['LFU', 'LFU-Lite', 'WLFU', 'LRU', 'LRUm', 'fLRU', 'CB-MPS', 'CB-SI', 'CB-SILite'],
      package_dir = {"":"cachingalgo"},
      include_package_data=True,
      package_data={
        "cachingalgo":["data/*.pkl"],
      },

      long_description = long_description,
      long_description_content_type="text/markdown",
      classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
      ],
      )
